#ssh_communicator.py
#========================================================================================================================================================================================================
#import psycopg2
#import pyodbc
import sys
from sshtunnel import SSHTunnelForwarder
import pandas as pd
from pprint import pprint
from pydantic import BaseModel, validator, create_model, Field
from typing import Optional, Dict
import re
import json

class DisplayConfig(BaseModel):
    ServiceId: str
    Model: str
    CustomerId: str
    ServiceName: str
    Description: Optional[str]
    AutonomousSystem: Optional[str]
    Interfaces: Optional[create_model('Interfaces', Interfaces=int, InterfaceName=dict)]
    #Static-Routing: Optional[create_model('Static-Routing', Entry=(dict))]
    #Interfaces: dict        #Optional[create_model('Interfaces', Interfaces=(int))]
    #var_name: dict = Field(alias='Static-Routing')
#class Config:
#   allow_population_by_field_name = True
class Interfaces(BaseModel):
    Interfaces: int
    InterfaceName: dict

class cSSH:
    def __init__(self,ip,jmp_ip=None):
        self.server = None
        self.local_bind_port = 5432
        self.ssh_ip = ip if jmp_ip is None else jmp_ip
        self.ssh_port = 22
        self.ssh_user = '<username>'
        self.ssh_password = '<password>'
        kwargs={'ssh_ip':self.ssh_ip,'ssh_port':self.ssh_port,'ssh_username':self.ssh_user,'ssh_password':self.ssh_password}
        self.open_ssh_tunnel(**kwargs)
        self.close_server()
    def open_ssh_tunnel(self,ssh_ip,ssh_port,ssh_username,ssh_password):    
        try:
            self.server = SSHTunnelForwarder(
                    (ssh_ip, ssh_port),
                    ssh_username=ssh_username,
                    ssh_password=ssh_password,
                    remote_bind_address=('localhost', self.local_bind_port))
            self.server.start()
            print("SSH server connected")
        except:
            print("SSH server not connected")
            sys.exit(0)
        if self.local_bind_port is not None:
            self.local_bind_port = str(self.server.local_bind_port)
    def close_server(self):
        self.server.close()
        print("SSH server closed")
class cDataProcessing():
    def __init__(self,path2file=None):
        self.file=path2file
        self.json_serviceusing=self.fct_data2json_serviceusing(self.file)
        self.json_displayconfig=self.fct_data2json_displayconfig(self.file)
        with open('serviceusing.json','w') as fp:
            json.dump(self.json_serviceusing,fp)
        with open('displayconfig.json','w') as fp:
            json.dump(self.json_displayconfig,fp)
        print("JSON-Files generated and saved.")
        #pprint(self.json_serviceusing)
        #pprint(self.json_displayconfig)
  # FUNCTIONS TO READ DATA
    def fct_data2json_serviceusing(self,path2file):
        df=pd.read_csv(path2file,sep="\t")
        lst_data=list(df.values)
        #print(df.head(20))
        lst_header=[]
        lst_data_service_using=[]
        is_data=False
        #for i in range(0,df.shape[0]):
        for i in range(0,len(lst_data)):
            if is_data==False:
                #if df.iloc[i][0][0:8]=='Services':
                if 'Services' in lst_data[i][0][0:8]:
                    is_data=True
            else:
                if lst_data[i][0][0:5]!='-----':
                    if lst_data[i][0][0:9]=='ServiceId':
                        lst_header=lst_data[i][0]
                    elif lst_data[i][0][0:5]!='=====':
                        lst_data_service_using.append(lst_data[i][0])
                elif lst_data[i][0][0:5]=='-----' and lst_data[i+2][0][0:5]=='-----':
                    is_data=False
        lst_header=list(filter(None, lst_header.split(' ')))
        lst_header[5]=lst_header[5] + lst_header[6]
        lst_header=lst_header[0:6]
        lst_data_service_using=[list(filter(None, x.split(' '))) for x in lst_data_service_using]
        lst_data_service_using_json=[]
        for x in lst_data_service_using:
            json={}
            for i in range(0,len(lst_header)):
                json.update({lst_header[i]:x[i]})
            lst_data_service_using_json.append(json)
        return lst_data_service_using_json
    def fct_rawdata_displayconfig(self,path2file):
        df=pd.read_csv(path2file,sep="\t")
        lst_data=list(df.values)
        is_data=False
        is_group=False
        dct_groups={}
        for i in range(0,len(lst_data)):
            if is_data==False:
                if 'echo "Service Configuration"' in lst_data[i][0][0:28]:
                    is_data=True
            else:
                row_act=lst_data[i][0]
                indent_act=len(row_act)-len(row_act.lstrip())
                if indent_act==8 and is_group==False and ('vprn' in lst_data[i][0] or 'vpls' in lst_data[i][0]):
                    row_bgn=i
                    group_key=lst_data[i][0]
                    group_key=group_key.replace("  ","")
                    is_group=True
                elif indent_act==8 and is_group==True:
                    row_end=i
                    is_group=False
                    lst_group_actual=lst_data[row_bgn:row_end]
                    if group_key not in dct_groups:
                        dct_groups.update({group_key:{'lst_rawdata':lst_group_actual}})
                    else:
                        if 'lst_rawdata' not in dct_groups[group_key]:
                            dct_groups[group_key].update({'lst_rawdata':lst_group_actual})
                        else:
                            dct_groups[group_key]['lst_rawdata'].extend(lst_group_actual[1:])
                    x=1
                x=1
        key_levels={}
        dct_json1={}
        multi_indent_error_key=[]
        for key in dct_groups.keys():
            val=dct_groups[key]
            lst_rawdata=val['lst_rawdata']
            dct_tmp={}
            for i in range(0,len(lst_rawdata)):
                act_string=lst_rawdata[i][0].replace(" create","")
                act_string_noident=act_string.lstrip()
                act_indent=len(act_string)-len(act_string_noident)
                if i<len(lst_rawdata)-1:
                    nxt_string=lst_rawdata[i+1][0]
                    nxt_indent=len(nxt_string)-len(nxt_string.lstrip())
                else:
                    nxt_string=""
                    nxt_indent=act_indent
                if act_indent-nxt_indent not in [-4,0,4]:
                    #print("Multi-Indent in Read-File!")
                    log=1
                if i==0:
                    act_key=act_string.replace("  ","")
                    key_levels=[act_key]
                    act_level=0
                    values=act_key.split(" ")
                    dct_json1.update({act_key:{}})
                    dct_json1[act_key].update({'Model':values[0].upper(),'ServiceId':values[1],'ServiceName':values[3],'CustomerId':values[5]})
                    dct_tmp=dct_json1[act_key]
                    x=1
                elif act_indent<nxt_indent:
                    act_level+=1
                    if nxt_indent-act_indent==4:
                        act_key=act_string.replace("  ","")
                        is_indent_error=False
                        jumps=0
                    else:
                        is_indent_error=True
                        jumps=int((nxt_indent-act_indent)/4)
                        for k in range(0,jumps-1):
                            act_key=f"indent_error_{len(multi_indent_error_key)+1}"
                            multi_indent_error_key.append(act_key)
                            key_levels.append(act_key)
                            dct_tmp.update({act_key:{}})
                            dct_tmp=dct_tmp[act_key]
                        act_level+=1                   
                        act_key=f"indent_error_{len(multi_indent_error_key)+1}"
                        multi_indent_error_key.append(act_key)
                        #key_levels.append(act_key)
                    key_levels.append(act_key)
                    dct_tmp.update({act_key:{}})
                    dct_tmp=dct_tmp[act_key]
                    x=1
                elif act_indent==nxt_indent:
                    dct_tmp.update({act_string.replace("  ",""):act_string.replace("  ","")})
                    x=1
                elif act_indent>nxt_indent:
                    dct_tmp.update({act_string.replace("  ",""):act_string.replace("  ","")})
                    key_levels.pop(len(key_levels)-1)
                    #if is_indent_error==True:
                    #    key_levels.pop(len(key_levels)-1)
                    act_level-=1
                    dct_tmp=dct_json1
                    for l in range(0,len(key_levels)):
                        dct_tmp=dct_tmp[key_levels[l]]
                    x=1
            x=2
        return dct_json1
    def fct_data2json_displayconfig(self,path2file):
        # Read data from file
        lst_json2=self.fct_rawdata_displayconfig(path2file)
        # Deepcopy of the dictionary to keep rawdata in lst_json2
        lst_json2b=self.fct_dct_copy(lst_json2)
        # Data was read with line-strings... Key and values are separated for the values in the following function (not yet for keys)
        self.update_nested_values(lst_json2b)
        # Not required keys are skipped in the following function
        lst_json2c=self.fct_dct_copy_specific(lst_json2b)
        # Build group for specific keys like Interfaces and BGP as required from the Data Model
        lst_json2d=self.fct_dct_build_groups(lst_json2c)
        # Transform the key-value pairs into the required data specification
        lst_json2e=self.fct_dct_clean_values(lst_json2d)
        # String values have been often saved with additional slashes "". These are removed in the following function.
        lst_json2f=self.fct_dct_remove_stringslashes(lst_json2e)
        return lst_json2f
  # FUNCTIONS TO TRANSFORM DATA INTO DICTIONARY/JSON-FORMAT
    def update_nested_values(self,in_dict):
        for k, v in in_dict.items():
            if isinstance(v, str):
                lst_values=v.split(" ")
                if v in ['no shutdown','shutdown']:
                    in_dict[k]=v
                elif len(lst_values)>1:
                    in_dict[k]=lst_values[1]
                else:
                    in_dict[k]=lst_values[0]
                #in_dict[k]=v.replace("  ","")
            elif isinstance(v, dict):
                self.update_nested_values(v)
            elif isinstance(v, list):
                for o in v:
                    if isinstance(o, dict):
                        self.update_nested_values(o)
    def fct_dct_copy(self,d):
        output = d.copy()
        for key, value in output.items():
            output[key] = self.fct_dct_copy(value) if isinstance(value, dict) else value        
        return output
    def fct_dct_copy_specific(self,d):
        lst_keys_final=['AutonomusSystem','ServiceId','Model','CustomerId','ServiceName','Description','Interfaces','InterfaceName','Status','IPv4Address','SapId','QosIngress','QoSEgress','VRRPId','VRRPBackup','VRRPPriority','BGP','Group','GroupName','Static-Routing','Entry','Next-Hop']    
        lst_keywords_include=['vprn','interface','description','autonomous-system','static-route-entry','next-hop','bgp','address','ipv6','sap','ingress','egress','qos','vrrp','vpls','backup','priority','policy','status','shutdown']
        lst_keywords_bgp=['group','family','peer-as','neighbor','import','export','keepalive','hold-time','next-hopself','type','local-as','as-override']
        lst_keywords_include.extend(lst_keywords_bgp)
        lst_keywords_exclude=['-bgp','vrf-']
        output = d.copy()
        for key, value in d.items():
            keyword_found=False
            for word in lst_keywords_include:
                if word in key:
                    keyword_found=True
            for word in lst_keywords_exclude:
                if word in key:
                    keyword_found=False
            if key in lst_keys_final:
                output[key] = self.fct_dct_copy_specific(value) if isinstance(value, dict) else value
            elif keyword_found==True:
                output[key] = self.fct_dct_copy_specific(value) if isinstance(value, dict) else value
            elif keyword_found==False:
                output.pop(key)
            elif isinstance(value,dict):
                output[key]=self.fct_dct_copy_specific(value)
            else:
                output.pop(key)
        return output
    def fct_dct_build_groups(self,d):
        dct_keyword2group={'interface':'Interfaces','static-route':'Static-Routing'}
        lst_keywords=[]
        for k in dct_keyword2group.keys():
            lst_keywords.append(k)
        output = self.fct_dct_copy(d)
        for k in d.keys():
            for key in d[k].keys():
                keyword_found=False
                for w in lst_keywords:
                    if w in key and keyword_found==False:
                        keyword_found=True
                        group_key=w
                if keyword_found==True:
                    group=dct_keyword2group[group_key]
                    if group in output[k]:
                        output[k][group].update({key:output[k][key]})
                    else:
                        output[k][group]={key:output[k][key]}
                    output[k].pop(key)
                    x=1
                elif keyword_found==False:
                    output[k][key]=d[k][key]
                    x=1
        return output
    def fct_dct_clean_values(self,d):
        lst_keywords_include=['description','autonomous-system','bgp','Static-Routing','Interfaces']
        dct_keyword2group={'description':'Description','autonomous-system':'AutonomousSystem','bgp':'BGP'}
        output = self.fct_dct_copy(d)
        for k in d.keys():
            for key in d[k].keys():
                keyword_found=False
                for w in lst_keywords_include:
                    if w in key and keyword_found==False:
                        keyword_found=True
                        key_string=w
                if keyword_found==True:
                    if key_string in ['description','autonomous-system']:
                        key_new=dct_keyword2group[key_string]
                        output[k][key_new]=output[k][key]
                        output[k].pop(key)
                    elif key_string in ['bgp']:
                        key_new=dct_keyword2group[key_string]
                        rawdata=output[k][key]
                        dct_bgp_groups={}
                        for bgp_key in rawdata.keys():
                            if 'group' in bgp_key:
                                groupID=bgp_key.split(" ")[1].replace('\"','')
                                groupLength=len(rawdata[bgp_key])
                                dct_bgp_groups[groupID]={'ID':groupID,'Length':groupLength}
                        val_group=0
                        val_groupname=[]
                        for gk in dct_bgp_groups.keys():
                            val_group+=dct_bgp_groups[gk]['Length']
                            newlist=[dct_bgp_groups[gk]['ID'] for i in range(0,dct_bgp_groups[gk]['Length'])]
                            val_groupname.append(newlist)
                        output[k][key_new]={'Group':val_group,'GroupName':val_groupname}
                        output[k].pop(key)
                    elif key_string in ['Static-Routing']:
                        key_new=key_string
                        rawdata=output[k][key]
                        dct_sr={'Entry':{}}
                        for sr_key in rawdata.keys():
                            if 'static-route-entry' in sr_key:
                                sr_entry=sr_key.split(" ")[1]
                                dct_sr['Entry'].update({sr_entry:{'Next-Hop':None}})
                                for sub_key in rawdata[sr_key].keys():
                                    if 'next-hop' in sub_key:
                                        sr_nexthop=sub_key.split(" ")[1]
                                        dct_sr['Entry'][sr_entry]['Next-Hop']=sr_nexthop
                        output[k][key].pop(sr_key)
                        output[k][key]=self.fct_dct_copy(dct_sr)
                    elif key_string in ['Interfaces']:
                        key_new=key_string
                        rawdata=output[k][key]
                        if_counter=0
                        dct_if={'Interfaces':None,'InterfaceName':{}}
                        if k[0:8]=='vprn 318':
                            x=1
                        for if_key in rawdata.keys():
                            if 'interface' in if_key:
                                if_counter+=1
                                if_name=if_key.split(" ")[1].replace('\"','')
                                if_ipv4_address, if_saplag, qos_egress, qos_ingress = None, None, None, None
                                if_vrrp_id, if_vrrp_backup, if_vrrp_priority = None, None, None
                                if_type=None
                                for sub_key in rawdata[if_key].keys():
                                    if 'address' in sub_key:
                                        if_ipv4_address=sub_key.split(" ")[1]
                                    if 'sap' in sub_key:
                                        if_saplag=sub_key.split(" ")[1]
                                        if_type='qos' if if_type is None else if_type
                                        for sub_key2 in rawdata[if_key][sub_key]:
                                            if 'ingress' in sub_key2:
                                                for sub_key3 in rawdata[if_key][sub_key][sub_key2].keys():
                                                    if 'qos' in sub_key3:
                                                        qos_ingress=sub_key3.split(" ")[1]
                                            if 'egress' in sub_key2:
                                                for sub_key3 in rawdata[if_key][sub_key][sub_key2].keys():
                                                    if 'qos' in sub_key3:
                                                        qos_egress=sub_key3.split(" ")[1]                           
                                    if 'vrrp' in sub_key:
                                        if_vrrp_id=sub_key.split(" ")[1]
                                        if_type='vrrp' if if_type is None else if_type
                                        for sub_key2 in rawdata[if_key][sub_key]:
                                            if 'backup' in sub_key2:
                                                if_vrrp_backup=sub_key2.split(" ")[1]
                                            if 'priority' in sub_key2:
                                                if_vrrp_priority=sub_key2.split(" ")[1]
                                    if 'vpls' in sub_key:
                                        if_sap_id=sub_key.split(" ")[1]
                                if if_type=='qos':
                                    dct_if['InterfaceName'].update({if_name:{'IPv4Address':if_ipv4_address
                                                            ,'SapId':{'SapId':if_saplag,'QosIngress':qos_ingress,'QosEgress':qos_egress}}})
                                elif if_type=='vrrp':
                                    dct_if['InterfaceName'].update({if_name:{'IPv4Address':{'IPv4Address':if_ipv4_address, 'VRRPId':if_vrrp_id, 'VRRPBackup':if_vrrp_backup, 'VRRPPriority':if_vrrp_priority}
                                                            ,'SapId':if_sap_id}})
                        dct_if['Interfaces']=if_counter
                        output[k][key]=self.fct_dct_copy(dct_if)
                        val_entry=1
                    x=1
            model=d[k]['Model']
            if model in ['vpls','VPLS']:
                dct_vpls_sap={}
                output[k]['SapId']={}
                for sub_key in d[k].keys():
                    if 'sap' in sub_key:
                        rawdata=output[k]
                        if_saplag=sub_key.split(" ")[1]
                        if_type='qos' if if_type is None else if_type
                        for sub_key2 in rawdata[sub_key].keys():
                            if 'ingress' in sub_key2:
                                for sub_key3 in rawdata[sub_key][sub_key2].keys():
                                    if 'qos' in sub_key3:
                                        qos_ingress=sub_key3.split(" ")[1]
                            if 'egress' in sub_key2:
                                for sub_key3 in rawdata[sub_key][sub_key2].keys():
                                    if 'qos' in sub_key3:
                                        qos_egress=sub_key3.split(" ")[1]                           
                        dct_vpls_sap.update({'SapId':if_saplag,'QosIngress':qos_ingress,'QosEgress':qos_egress})
                        output[k]['SapId'].update({if_saplag:self.fct_dct_copy(dct_vpls_sap)})
                        output[k].pop(sub_key)
        return output
    def fct_dct_remove_stringslashes(self,d):
        output = d.copy()
        for key, value in output.items():
            if isinstance(value, dict):
                output[key] = self.fct_dct_remove_stringslashes(value) 
            elif isinstance(value,str):
                output[key] = value.replace('\"','')
            else:
                output[key] = value
        return output

def fct_re_ipcheck(ip):
    re_ip=re.compile("^\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}$")
    match_ip=re.match(re_ip,ip)
    check_ip=False if match_ip is None else True
    return check_ip
def fct_runSSH(ssh_ip=None,jump_ip=None):
  # CONNECT SSH
    #sys.argv=['ssh_communicator.py', 'invalid ip-string', '-jmp', '100.100.0.0']
    for i in range(0,len(sys.argv)):
        found_ip=False
        if isinstance(sys.argv[i],str) and i<=1 and found_ip==False:
            if fct_re_ipcheck(sys.argv[i]):
                ssh_ip=sys.argv[i]
                found_ip==True
        if sys.argv[i] in ['jmp','-jmp']:
            jump_ip=sys.argv[i+1] if i+1<len(sys.argv) else jump_ip
    check_ip=fct_re_ipcheck(ssh_ip) if ssh_ip is not None else False
    check_jump_ip=fct_re_ipcheck(jump_ip) if jump_ip is not None else False
    SSH_connect=False
    if check_ip==True and jump_ip is None:
        clsSSH=cSSH(ssh_ip)
        SSH_connect=True
    elif check_jump_ip==True:
        clsSSH=cSSH(ssh_ip,jump_ip)
        SSH_connect=True
    elif check_ip==False and jump_ip is None:
        raise Exception("invalid SSH ip-address")
    elif check_jump_ip ==False:
        raise Exception("invalid jump-ip provided")
    else:
        raise Exception("SSH-connection could not have been established")
    #x=input("Press Enter")
    return SSH_connect

if __name__=='__main__':
    print(sys.argv)
    try:
        fct_runSSH()
    except Exception as e:
        print(f"Exception occurred: '{e}'")
        pass
  # READ DATA FROM LOG-FILE
    #path2file=r'C:\Users\BG\OneDrive\BG\coding\3_Productive\Python\adhoc\SBB' + chr(92)
    #clsDP=cDataProcessing(path2file+"input_data.log")
    clsDP=cDataProcessing("input_data.log")
  # DATA VALIDATION
    for key in clsDP.json_displayconfig.keys():
        validate=DisplayConfig.parse_obj(clsDP.json_displayconfig[key])
        #print(validate)
    stophere=1


    